/*
 * S I L E N C E
 *
 * An auto-extensible system for making music on computers by means of
 * software alone.
 * Copyright (c) 2001 by Michael Gogins. All rights reserved.
 *
 * L I C E N S E
 *
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * P U R P O S E
 *
 * Defines the public C application programming interface to Csound.
 *
 * 29 May 2002 - ma++ merge with CsoundLib.
 * 30 May 2002 - mkg add csound "this" pointer argument back into merge.
 * 27 Jun 2002 - mkg complete Linux dl code and Makefile
 * 30 Jan 2003 - IV rewrote most of this file
 */
#ifdef __cplusplus
extern "C" {
#endif

#include <cs.h>
#include "timers.h"

#ifndef MAX_PATH
static const int MAX_PATH = 255;
#endif

/*
 * Csound symbols referenced in this file.
 */

#define csoundMaxExits 64

static  void    *csoundExitFuncs_[csoundMaxExits];
static  long    csoundNumExits_ = -1;
static  jmp_buf csoundJump_;

extern  int     csoundMain(void *csound, int argc, char **argv);
extern  void    csoundMainCleanup(void);
extern  void    mainRESET(void);
extern  void    dispkill(WINDAT *windat);
extern  long    kperf(long kcnt);
extern  long    kcnt;
extern  char    *inbuf;
extern  void    *outbuf;
extern  void    SfReset(void);
extern  void    csoundDefaultMidiOpen(void *csound);
extern  void    create_opcodlst(void);
extern  int     writeLine(const char *text, long size);
extern  void    newevent(/* void *csound, */ char type, MYFLT *pfields,
                         long numFields);
extern  opcodelist  *new_opcode_list(void);
extern  int     dispose_opcode_list(opcodelist *list);
extern  void    csoundDefaultExternalMidiCloseCallback(void *csound);

static  int     instance_cnt = 0;

/* for measuring elapsed time */

static RTCLOCK clk_bas = { 0UL, 0UL, 0.0, 0UL, 0UL, 0.0 };

/* This needs a stub definition in this file. */

char *getDB(void)
{
    return NULL;
}

void *csoundCreate(void *hostData)
{
    /* Currently, there is only one instance of Csound per process. */
    /* In the future, this will be dynamically allocated, and will  */
    /* need to be passed to all internals Csound functions that     */
    /* reference it. */
    if (instance_cnt) {
      csoundMessage(NULL, "This CsoundAPI implementation does not support "
                          "multiple instances\n");
      return NULL;
    }
    instance_cnt++;
    globals_init();
    cglob_init();
    csoundReset(NULL);
    cglob.hostdata = hostData;
    timers_struct_init(&clk_bas);       /* start timer */
    return &cglob;
}

int csoundQueryInterface(const char *name, void **interface, int *version)
{
    if(strcmp(name, "CSOUND") == 0)
    {
      *interface = csoundCreate(NULL);
      *version = csoundGetVersion();
      return 0;
    }
    return 1;
}

void csoundDestroy(void *csound)
{
    IGN(csound);
    if (!instance_cnt)
      csoundMessage(NULL, "No Csound instance to destroy.\n");
    else {
      csoundReset(NULL);
      --instance_cnt;
    }
}

int csoundGetVersion(void)
{
    /* should not there be a way to allow patchlevels ? */
    return (VERSION * 100 + SUBVER);
}

int csoundGetAPIVersion(void)
{
    return (APIVERSION * 100 + APISUBVER);
}

void *csoundGetHostData(void *csound)
{
    IGN(csound);
    return cglob.hostdata;
}

void csoundSetHostData(void *csound, void *hostData)
{
    IGN(csound);
    cglob.hostdata = hostData;
}

/*
 * PERFORMANCE
 */

extern  int     runincomponents;
extern  int     sensevents(void);
extern  int     cleanup(void);
extern  int     orcompact(void);

int csoundPerform(void *csound, int argc, char **argv)
{
    int returnValue;
    IGN(csound);
    /* do we have a Csound instance to perform ? */
    if (!instance_cnt) {
      err_printf("error: csoundPerform was called without csoundCreate\n");
      return CSOUND_ERROR;
    }
    runincomponents = 0;
    returnValue = csoundMain(NULL, argc, argv);
    if (returnValue) {
      err_printf("Early return from csoundPerform().\n");
    }
    return returnValue;
}

int csoundCompile(void *csound, int argc, char **argv)
{
    int returnValue;
    IGN(csound);
    /* do we have a Csound instance to perform ? */
    if (!instance_cnt) {
      err_printf("error: csoundCompile was called without csoundCreate\n");
      return CSOUND_ERROR;
    }
    runincomponents = 1;
    returnValue = csoundMain(NULL, argc, argv);
    if (returnValue) {
      err_printf("Early return from csoundCompile().\n");
    }
    return returnValue;
}

int csoundPerformKsmps(void *csound)
{
    int done = 0;
    volatile int returnValue;
    IGN(csound);
    /* setup jmp for return after an exit() */
    if ((returnValue = setjmp(csoundJump_)) != 0) {
      csoundMessage(NULL, "Early return from csoundPerformKsmps().\n");
      return returnValue;
    }
    done = sensevents();
    if (!done && kcnt)
    {
      /*
        Rather than overriding real-time event handling in kperf,
        turn it off before calling kperf, and back on afterwards.
       */
      int rtEvents = CSGLOBAL(O).RTevents;
      CSGLOBAL(O).RTevents = 0;
      kperf(1);
      kcnt -= 1;
      CSGLOBAL(O).RTevents = rtEvents;
    }
    return done;
}

/* external host's outbuffer passed in csoundPerformBuffer() */

#if 0
static  char    *_rtCurOutBuf       = NULL;
static  long    _rtCurOutBufSize    = 0;
static  long    _rtCurOutBufCount   = 0;
static  char    *_rtOutOverBuf      = NULL;
static  long    _rtOutOverBufSize   = 0;
static  long    _rtOutOverBufCount  = 0;
static  char    *_rtInputBuf        = NULL;
static  long    _rtInputBufSize     = 0;
static  long    _rtInputBufIndex    = 0;
#endif

int csoundPerformBuffer(void *csound)
{
    IGN(csound);
    printf("csoundPerformBuffer(): DO NOT USE THIS INTERFACE\n");
    return(-1);
#if 0
    volatile int returnValue;
    /* Number of samples still needed to create before returning. */
    static int sampsNeeded = 0;
    int sampsPerKperf = ksmps * nchnls;
    int done = 0;
    /* Setup jmp for return after an exit().
    */
    if (returnValue = setjmp(csoundJump_)) {
      csoundMessage(csound, "Early return from csoundPerformBuffer().");
      return returnValue;
    }
    _rtCurOutBufCount = 0;
    sampsNeeded += CSGLOBAL(O).outbufsamps;
    while (!done && sampsNeeded > 0) {
      done = sensevents();
      if (done) {
        return done;
      }
      if (kcnt) {
        int rtEvents = CSGLOBAL(O).RTevents;
        CSGLOBAL(O).RTevents = 0;
        kperf(1);
        kcnt -= 1;
        sampsNeeded -= sampsPerKperf;
        CSGLOBAL(O).RTevents = rtEvents;
      }
    }
    return done;
#endif
}

void csoundCleanup(void *csound)
{
    void remove_tmpfiles(void);
    IGN(csound);
    orcompact();
    cleanup();
    remove_tmpfiles();
    /* Call all the funcs registered with atexit(). */
    while (csoundNumExits_ >= 0)
    {
      void (*func)(void) = (void (*)(void)) csoundExitFuncs_[csoundNumExits_];
      func();
      csoundNumExits_--;
    }
}

/*
 * ATTRIBUTES
 */

MYFLT csoundGetSr(void *csound)                 /* sample rate */
{
    IGN(csound);
    return esr;
}

MYFLT csoundGetKr(void *csound)                 /* control rate of current */
{                                               /* instrument instance */
    IGN(csound);
    return ekr;
}

MYFLT csoundGetGlobalKr(void *csound)           /* orchestra level */
{                                               /* control rate */
    IGN(csound);
    return global_ekr;
}

int csoundGetKsmps(void *csound)                /* ksmps of current */
{                                               /* instrument instance */
    IGN(csound);
    return (int) ksmps;
}

int csoundGetGlobalKsmps(void *csound)          /* orchestra level ksmps */
{
    IGN(csound);
    return (int) global_ksmps;
}

int csoundGetNchnls(void *csound)               /* number of channels */
{
    IGN(csound);
    return (int) nchnls;       /* should we assume input is same as output ? */
}

MYFLT csoundGet0dBFS(void *csound)              /* 0dBFS level */
{
    IGN(csound);
    return e0dbfs;
}

int csoundGetSampleFormat(void *csound)         /* output sample format */
{
    IGN(csound);
    /* should we assume input is same as output ? */
    return (int) CSGLOBAL(O).outformat;
}

int csoundGetSampleSize(void *csound)           /* output sample size */
{
    IGN(csound);
    /* should we assume input is same as output ? */
    return (int) CSGLOBAL(O).outsampsiz;
}

long csoundGetInputBufferSize(void *csound)     /* input buffer size */
{                                               /* multiplied by nchnls */
    IGN(csound);
    return (long) CSGLOBAL(O).inbufsamps;
}

long csoundGetOutputBufferSize(void *csound)    /* output buffer size */
{                                               /* multiplied by nchnls */
    IGN(csound);
    return (long) CSGLOBAL(O).outbufsamps;
}

void *csoundGetInputBuffer(void *csound)        /* return pointer to */
{                                               /* input buffer */
    IGN(csound);                                /* (not implemented yet) */
    return NULL;
}

void *csoundGetOutputBuffer(void *csound)       /* return pointer to */
{                                               /* output buffer */
    IGN(csound);
    return outbuf;
}

MYFLT *csoundGetSpin(void *csound)
{
    IGN(csound);
    return spin;
}

MYFLT *csoundGetSpout(void *csound)
{
    IGN(csound);
    return spout;
}

MYFLT csoundGetScoreTime(void *csound)
{
    IGN(csound);
    /* is it expected to be in seconds ? The documentation is not clear ... */
    return (MYFLT) ((double) global_kcounter * (double) global_onedkr);
}

MYFLT csoundGetProgress(void *csound)
{
    IGN(csound);
    return FL(0.0);     /* do not know */
}

/* score time / real time */

MYFLT csoundGetProfile(void *csound)
{
    MYFLT t;
    IGN(csound);
    t = (MYFLT) timers_get_real_time(&clk_bas);
    if (t <= FL(0.0))
      return (FL(1.0));
    t = (MYFLT) ((double) global_kcounter * (double) global_onedkr) / t;
    return t;
}

/* CPU time / real time */

MYFLT csoundGetCpuUsage(void *csound)
{
    MYFLT t;
    IGN(csound);
    t = (MYFLT) timers_get_real_time(&clk_bas);
    if (t <= FL(0.0))
      return (FL(1.0));
    t = (MYFLT) timers_get_CPU_time(&clk_bas) / t;
    return t;
}

/*
 * SCORE HANDLING
 */

static int csoundIsScorePending_ = 0;

int csoundIsScorePending(void *csound)
{
    IGN(csound);
    return csoundIsScorePending_;
}

void csoundSetScorePending(void *csound, int pending)
{
    IGN(csound);
    csoundIsScorePending_ = pending;
}

static MYFLT csoundScoreOffsetSeconds_ = (MYFLT) 0.0;

void csoundSetScoreOffsetSeconds(void *csound, MYFLT offset)
{
    IGN(csound);
    csoundScoreOffsetSeconds_ = offset;
}

MYFLT csoundGetScoreOffsetSeconds(void *csound)
{
    IGN(csound);
    return csoundScoreOffsetSeconds_;
}

void csoundRewindScore(void *csound)
{
    IGN(csound);
    if(scfp) {
      fseek(scfp, 0, SEEK_SET);
    }
}

static void csoundDefaultMessageVCallback(void *csound, const char *format,
                                          va_list args)
{
    IGN(csound);
    vfprintf(stderr, format, args);
}

static void csoundDefaultMessageSCallback(void *csound, const char *format,
                                          va_list args)
{
    IGN(csound);
    vfprintf(stdout, format, args);
}

static void (*csoundMessageVCallback_)(void *csound, const char *format,
                                       va_list args) =
    csoundDefaultMessageVCallback;

static void (*csoundMessageSCallback_)(void *csound, const char *format,
                                       va_list args) =
    csoundDefaultMessageSCallback;

/* set both message callbacks */

void csoundSetMessageCallback(void *csound,
                              void (*csoundMessageCallback)
                                   (void *csound, const char *format,
                                    va_list args))
{
    IGN(csound);
    csoundMessageVCallback_ = csoundMessageCallback;
    csoundMessageSCallback_ = csoundMessageCallback;
}

/* set stderr callback only */

void csoundSetMessageVCallback(void *csound,
                               void (*csoundMessageCallback)
                                    (void *csound, const char *format,
                                     va_list args))
{
    IGN(csound);
    csoundMessageVCallback_ = csoundMessageCallback;
}

/* set stdout callback only */

void csoundSetMessageSCallback(void *csound,
                               void (*csoundMessageCallback)
                                    (void *csound, const char *format,
                                     va_list args))
{
    IGN(csound);
    csoundMessageSCallback_ = csoundMessageCallback;
}

void csoundMessageV(void *csound, const char *format, va_list args)
{
    csoundMessageVCallback_(csound, format, args);
}

void csoundMessageS(void *csound, const char *format, va_list args)
{
    csoundMessageSCallback_(csound, format, args);
}

void csoundMessage(void *csound, const char *format, ...)
{
    va_list args;
    va_start(args, format);
    csoundMessageSCallback_(csound, format, args);
    va_end(args);
}

/* for compatibility */

void dribble_printf(const char *format, ...)
{
    va_list args;
    if (dribble != NULL) {
      va_start(args, format);
      vfprintf(dribble, format, args);
      va_end(args);
    }
    va_start(args, format);
    csoundMessageSCallback_(&cglob, format, args);
    va_end(args);
}

void err_printf(const char *format, ...)
{
    va_list args;
    va_start(args, format);
    csoundMessageVCallback_(&cglob, format, args);
    va_end(args);
    if (dribble != NULL) {
      va_start(args, format);
      vfprintf(dribble, format, args);
      va_end(args);
    }
}

/* not sure what these do */

static void (*csoundThrowMessageCallback_)(void *csound, const char *format,
                                           va_list args) =
    csoundDefaultMessageVCallback;

void csoundSetThrowMessageCallback(void *csound,
                                   void (*csoundThrowMessageCallback)
                                        (void *csound, const char *format,
                                         va_list args))
{
    IGN(csound);
    csoundThrowMessageCallback_ = csoundThrowMessageCallback;
}

void csoundThrowMessageV(void *csound, const char *format, va_list args)
{
    csoundThrowMessageCallback_(csound, format, args);
}

void csoundThrowMessage(void *csound, const char *format, ...)
{
    va_list args;
    va_start(args, format);
    csoundThrowMessageCallback_(csound, format, args);
    va_end(args);
}

/* get / set message level */

int csoundGetMessageLevel(void *csound)
{
    IGN(csound);
    return CSGLOBAL(O).msglevel;
}

void csoundSetMessageLevel(void *csound, int messageLevel)
{
    IGN(csound);
    CSGLOBAL(O).msglevel = messageLevel;
}

void csoundInputMessage(void *csound, const char *message)
{
    IGN(csound);
    writeLine(message, strlen(message));
}

static char inChar_ = 0;

void csoundKeyPress(void *csound, char c)
{
    IGN(csound);
    inChar_ = c;
}

char getChar(void)
{
    return inChar_;
}

  /*
   * CONTROL AND EVENTS
   */

static void (*csoundInputValueCallback_)(void *csound, char *channelName,
                                         MYFLT *value) = NULL;

void csoundSetInputValueCallback(void *csound,
                                 void (*inputValueCalback)
                                      (void *csound, char *channelName,
                                       MYFLT *value))
{
    IGN(csound);
    csoundInputValueCallback_ = inputValueCalback;
}

void InputValue(char *channelName, MYFLT *value)
{
    if (csoundInputValueCallback_) {
      csoundInputValueCallback_(&cglob, channelName, value);
    }
    else {
      *value = FL(0.0);
    }
}

static void (*csoundOutputValueCallback_)(void *csound, char *channelName,
                                          MYFLT value) = NULL;

void csoundSetOutputValueCallback(void *csound,
                                  void (*outputValueCalback)
                                       (void *csound, char *channelName,
                                        MYFLT value))
{
    IGN(csound);
    csoundOutputValueCallback_ = outputValueCalback;
}

void OutputValue(char *channelName, MYFLT value)
{
    if (csoundOutputValueCallback_) {
      csoundOutputValueCallback_(&cglob, channelName, value);
    }
}

void csoundScoreEvent(void *csound, char type, MYFLT *pfields, long numFields)
{
    IGN(csound);
    newevent(type, pfields, numFields);
}

  /*
   * REAL-TIME AUDIO
   */

#ifdef RTAUDIO

extern void playopen_(int nchanls, int dsize, float sr, int scale);

void (*playopen)(int nchanls, int dsize, float sr, int scale) = playopen_;

/* open for audio output */
#if 0
void playopen_mi(int nchanls, int dsize, float sr, int scale)
{
    _rtCurOutBufSize = CSGLOBAL(O).outbufsamps * dsize;
    _rtCurOutBufCount = 0;
    _rtCurOutBuf = mmalloc(_rtCurOutBufSize);

    /* a special case we need to handle 'overlaps'
    */
    if (ksmps * nchanls > CSGLOBAL(O).outbufsamps)
    {
      _rtOutOverBufSize = (ksmps*nchanls - CSGLOBAL(O).outbufsamps) * dsize;
      _rtOutOverBuf = mmalloc(_rtOutOverBufSize);
      _rtOutOverBufCount = 0;
    }
    else
    {
      _rtOutOverBufSize = 0;
      _rtOutOverBuf = 0;
      _rtOutOverBufCount = 0;
    }
}
#endif

extern void rtplay_(char *outBuf, int nbytes);

void (*rtplay)(char *outBuf, int nbytes) = rtplay_;

#if 0
void rtplay_mi(char *outBuf, int nbytes)
{
    int bytes2copy = nbytes;
    /* copy any remaining samps from last buffer
     */
    if (_rtOutOverBufCount)
    {
      memcpy(_rtCurOutBuf, _rtOutOverBuf, _rtOutOverBufCount);
      _rtCurOutBufCount = _rtOutOverBufCount;
      _rtOutOverBufCount = 0;
    }
    /* handle any new 'overlaps'
     */
    if (bytes2copy + _rtCurOutBufCount > _rtCurOutBufSize)
    {
      _rtOutOverBufCount = _rtCurOutBufSize - (bytes2copy + _rtCurOutBufCount);
      bytes2copy = _rtCurOutBufSize - _rtCurOutBufCount;

      memcpy(_rtOutOverBuf, outBuf+bytes2copy, _rtOutOverBufCount);
    }
    /* finally copy the buffer
     */
    memcpy(_rtCurOutBuf+_rtOutOverBufCount, outBuf, bytes2copy);
    _rtCurOutBufCount += bytes2copy;
}
#endif

extern void recopen_(int nchanls, int dsize, float sr, int scale);

void (*recopen)(int nchanls, int dsize, float sr, int scale) = recopen_;

#if 0
void recopen_mi(int nchanls, int dsize, float sr, int scale)
{
    if (CSGLOBAL(O).inbufsamps*CSGLOBAL(O).outsampsiz !=
        CSGLOBAL(O).outbufsamps*CSGLOBAL(O).insampsiz)
      die("Input buffer must be the same size as Output buffer\n");
    _rtInputBuf = mmalloc(CSGLOBAL(O).inbufsamps*CSGLOBAL(O).insampsiz);
}
#endif

extern int rtrecord_(char *inBuf, int nbytes);

int (*rtrecord)(char *inBuf, int nbytes) = rtrecord_;

#if 0
int rtrecord_mi(char *inBuf, int nbytes)
{
    memcpy(inBuf, _rtInputBuf, nbytes);
    return nbytes;
}
#endif

extern void rtclose_(void);

void (*rtclose)(void) = rtclose_;

#if 0
void rtclose_mi(void)
{
    if (_rtCurOutBuf)
      mfree(_rtCurOutBuf);
    if (_rtOutOverBuf)
      mfree(_rtOutOverBuf);
    if (_rtInputBuf)
      mfree(_rtInputBuf);
}
#endif

#endif

void csoundSetPlayopenCallback(void *csound,
                               void (*playopen__)(int nchanls, int dsize,
                                                  float sr, int scale))
{
    IGN(csound);
#ifdef RTAUDIO
    playopen = playopen__;
#endif
}

void csoundSetRtplayCallback(void *csound,
                             void (*rtplay__)(char *outBuf, int nbytes))
{
    IGN(csound);
#ifdef RTAUDIO
    rtplay = rtplay__;
#endif
}

void csoundSetRecopenCallback(void *csound,
                              void (*recopen__)(int nchanls, int dsize,
                                                float sr, int scale))
{
    IGN(csound);
#ifdef RTAUDIO
    recopen = recopen__;
#endif
}

void csoundSetRtrecordCallback(void *csound,
                               int (*rtrecord__)(char *inBuf, int nbytes))
{
    IGN(csound);
#ifdef RTAUDIO
    rtrecord = rtrecord__;
#endif
}

void csoundSetRtcloseCallback(void *csound, void (*rtclose__)(void))
{
    IGN(csound);
#ifdef RTAUDIO
    rtclose = rtclose__;
#endif
}

  /*
   * MIDI -- this code replaces the MIDIDevice.c file
   */

static void (*csoundExternalMidiOpenCallback_)(void *csound) =
    csoundDefaultMidiOpen;

void csoundSetExternalMidiOpenCallback(void *csound,
                                       void (*csoundMidiOpen)(void *csound))
{
    IGN(csound);
    csoundExternalMidiOpenCallback_ = csoundMidiOpen;
}

void csoundExternalMidiOpen(void *csound)
{
    csoundExternalMidiOpenCallback_(csound);
}

static int defaultCsoundMidiRead(void *csound, unsigned char *midiData,
                                 int size)
{
    IGN(csound);
    IGN(midiData);
    IGN(size);
    return 0;
}

static int (*csoundExternalMidiReadCallback_)(void *csound,
                                              unsigned char *midiData,
                                              int size) =
    defaultCsoundMidiRead;

void csoundSetExternalMidiReadCallback(void *csound,
                                       int (*midiReadCallback)
                                           (void *csound,
                                            unsigned char *midiData, int size))
{
    IGN(csound);
    csoundExternalMidiReadCallback_ = midiReadCallback;
}

int csoundExternalMidiRead(void *csound, unsigned char *mbuf, int size)
{
    return csoundExternalMidiReadCallback_(csound, mbuf, size);
}

static int csoundExternalMidiEnabled_ = 0;

int csoundIsExternalMidiEnabled(void *csound)
{
    IGN(csound);
    return csoundExternalMidiEnabled_;
}

void csoundSetExternalMidiEnabled(void *csound, int enabled)
{
    IGN(csound);
    csoundExternalMidiEnabled_ = enabled;
}

static int defaultCsoundMidiWrite(void *csound, unsigned char *midiData)
{
    IGN(csound);
    IGN(midiData);
    return 0;
}

static int (*csoundExternalMidiWriteCallback_)(void *csound,
                                               unsigned char *midiData) =
    defaultCsoundMidiWrite;

void csoundSetExternalMidiWriteCallback(void *csound,
                                        int (*midiWriteCallback)
                                            (void *csound,
                                             unsigned char *midiData))
{
    IGN(csound);
    csoundExternalMidiWriteCallback_ = midiWriteCallback;
}

int csoundExternalMidiWrite(unsigned char *midiData)
{
    return csoundExternalMidiWriteCallback_(&cglob, midiData);
}

void csoundDefaultMidiCloseCallback(void *csound)
{
    IGN(csound);
}

static void (*csoundExternalMidiCloseCallback_)(void *csound) =
    csoundDefaultMidiCloseCallback;

void csoundSetExternalMidiCloseCallback(void *csound,
                                        void (*csoundExternalMidiCloseCallback)
                                             (void *csound))
{
    IGN(csound);
    csoundExternalMidiCloseCallback_ = csoundExternalMidiCloseCallback;
}

void csoundExternalMidiClose(void *csound)
{
    csoundExternalMidiCloseCallback_(csound);
}

  /*
   * FUNCTION TABLE DISPLAY.
   */

#ifdef WINDOWS
extern  int     Graphable_(void);
extern  void    MakeGraph_(WINDAT *, char *);
extern  void    DrawGraph_(WINDAT *);
extern  void    KillGraph_(WINDAT *);
extern  int     ExitGraph_(void);
extern  void    MakeXYin_(XYINDAT *, MYFLT, MYFLT);
extern  void    ReadXYin_(XYINDAT *);
#endif
extern  void    MakeAscii(WINDAT *, char *);
extern  void    DrawAscii(WINDAT *);
extern  void    KillAscii(WINDAT *);

static int isGraphable_ = 1;

void csoundSetIsGraphable(void *csound, int isGraphable)
{
    IGN(csound);
    isGraphable_ = isGraphable;
}

int Graphable(void)
{
#ifdef WINDOWS
    return Graphable_();
#else
    return isGraphable_;
#endif
}

static void defaultCsoundMakeGraph(void *csound, WINDAT *windat, char *name)
{
    IGN(csound);
#ifdef WINDOWS
    MakeGraph_(windat, name);
#else
    MakeAscii(windat, name);
#endif
}

static void (*csoundMakeGraphCallback_)(void *csound,  WINDAT *windat,
                                        char *name) =
    defaultCsoundMakeGraph;

void csoundSetMakeGraphCallback(void *csound,
                                void (*makeGraphCallback)
                                     (void *csound, WINDAT *windat,
                                      char *name))
{
    IGN(csound);
    csoundMakeGraphCallback_ = makeGraphCallback;
}

void MakeGraph(WINDAT *windat, char *name)
{
    csoundMakeGraphCallback_(&cglob, windat, name);
}

static void defaultCsoundDrawGraph(void *csound, WINDAT *windat)
{
    IGN(csound);
#ifdef WINDOWS
    DrawGraph_(windat);
#else
    DrawAscii(windat);
#endif
}

static void (*csoundDrawGraphCallback_)(void *csound,  WINDAT *windat) =
    defaultCsoundDrawGraph;

void csoundSetDrawGraphCallback(void *csound,
                                void (*drawGraphCallback)(void *csound,
                                                          WINDAT *windat))
{
    IGN(csound);
    csoundDrawGraphCallback_ = drawGraphCallback;
}

void DrawGraph(WINDAT *windat)
{
    csoundDrawGraphCallback_(&cglob, windat);
}

static void defaultCsoundKillGraph(void *csound, WINDAT *windat)
{
    IGN(csound);
#ifdef WINDOWS
    KillGraph_(windat);
#else
    KillAscii(windat);
#endif
}

static void (*csoundKillGraphCallback_)(void *csound, WINDAT *windat) =
    defaultCsoundKillGraph;

void csoundSetKillGraphCallback(void *csound,
                                void (*killGraphCallback)(void *csound,
                                                          WINDAT *windat))
{
    IGN(csound);
    csoundKillGraphCallback_ = killGraphCallback;
}

void KillGraph(WINDAT *windat)
{
    csoundKillGraphCallback_(&cglob, windat);
}

static int defaultCsoundExitGraph(void *csound)
{
    IGN(csound);
    return CSOUND_SUCCESS;
}

static int (*csoundExitGraphCallback_)(void *csound) = defaultCsoundExitGraph;

void csoundSetExitGraphCallback(void *csound,
                                int (*exitGraphCallback)(void *csound))
{
    IGN(csound);
    csoundExitGraphCallback_ = exitGraphCallback;
}

int ExitGraph(void)
{
    return csoundExitGraphCallback_(&cglob);
}

void MakeXYin(XYINDAT *xyindat, MYFLT x, MYFLT y)
{
#ifdef WINDOWS
    MakeXYin_(xyindat, x, y);
#else
    printf("xyin not supported. use invalue opcode instead.\n");
#endif
}

void ReadXYin(XYINDAT *xyindat)
{
#ifdef WINDOWS
    ReadXYin_(xyindat);
#else
    printf("xyin not supported. use invlaue opcodes instead.\n");
#endif
}

  /*
   * OPCODES
   */

opcodelist *csoundNewOpcodeList()
{
    create_opcodlst();
    return new_opcode_list();
}

void csoundDisposeOpcodeList(opcodelist *list)
{
    dispose_opcode_list(list);
}

int csoundAppendOpcode(char *opname,
                       int  dsblksiz,
                       int  thread,
                       char *outypes,
                       char *intypes,
                       SUBR iopadr,
                       SUBR kopadr,
                       SUBR aopadr,
                       SUBR dopadr)
{
    int oldSize = (int)((char *)oplstend - (char *)opcodlst);
    int newSize = oldSize + sizeof(OENTRY);
    int oldCount = oldSize / sizeof(OENTRY);
    int newCount = oldCount + 1;
    OENTRY *oldOpcodlst = opcodlst;
    opcodlst = (OENTRY *) mrealloc(opcodlst, newSize);
    if (!opcodlst) {
      opcodlst = oldOpcodlst;
      err_printf("Failed to allocate new opcode entry.");
      return 0;
    }
    else {
      OENTRY *oentry = opcodlst + oldCount;
      oplstend = opcodlst + newCount;
      oentry->opname    = opname;
      oentry->dsblksiz  = dsblksiz;
      oentry->thread    = thread;
      oentry->outypes   = outypes;
      oentry->intypes   = intypes;
      oentry->iopadr    = iopadr;
      oentry->kopadr    = kopadr;
      oentry->aopadr    = aopadr;
      oentry->dopadr    = dopadr;
      printf("Appended opcodlst[%d]: opcode = %s intypes = %s outypes = %s\n",
        oldCount,
        oentry->opname,
        oentry->intypes,
        oentry->outypes);
      return 1;
    }
}

int csoundOpcodeCompare(const void *v1, const void *v2)
{
    return strcmp(((OENTRY*)v1)->opname, ((OENTRY*)v2)->opname);
}

void csoundOpcodeDeinitialize(void *csound, INSDS *ip_)
{
    OPDS *p;
    IGN(csound);
    /* FIXME: this code has no protection against de-initialising */
    /* the same opcode twice (both in init and perf chain) */
    p = (OPDS*) ip_->nxti;
    while (p) {
      if (p->dopadr) {
        (*p->dopadr)(p);
      }
      p = p->nxti;
    }
    p = (OPDS*) ip_->nxtp;
    while (p) {
      if (p->dopadr) {
        (*p->dopadr)(p);
      }
      p = p->nxtp;
    }
}

  /*
   * MISC FUNCTIONS
   */

  /*
   * to enable csoundYield(), you must define in cs.h:
   * #define POLL_EVENTS() csoundYield()
   */

static int defaultCsoundYield(void *csound)
{
    IGN(csound);
    return 1;
}

static int (*csoundYieldCallback_)(void *csound) = defaultCsoundYield;

void csoundSetYieldCallback(void *csound, int (*yieldCallback)(void *csound))
{
    IGN(csound);
    csoundYieldCallback_ = yieldCallback;
}

int csoundYield(void *csound)
{
    return csoundYieldCallback_(csound);
}

static const int MAX_ENVIRONS = 128;

typedef struct Environs
{
    char *name;
    char *value;
} Environs;

static Environs *csoundEnv_ = NULL;

static int csoundNumEnvs_ = 0;

void csoundSetEnv(void *csound, const char *name, const char *value)
{
    int i = 0, l;
    IGN(csound);
    if (!name || !value)
      return;

    if (csoundEnv_ == NULL) {
      csoundEnv_ = (Environs *) mcalloc(MAX_ENVIRONS * sizeof(Environs));
    }
    l = strlen(value) + 1;
    for (i = 0; i < csoundNumEnvs_; i++) {
      if (strcmp(csoundEnv_[i].name, name) == 0) {
        mrealloc(csoundEnv_[i].value, l);
        strcpy(csoundEnv_[i].value, value);
        return;
      }
    }
    if (i >= MAX_ENVIRONS) {
      die("Exceeded maximum number of environment values");
      return;
    }

    csoundNumEnvs_++;
    csoundEnv_[i].name = mmalloc(strlen(name) + 1);
    strcpy(csoundEnv_[i].name, name);
    csoundEnv_[i].value = mmalloc(l);
    strcpy(csoundEnv_[i].value, value);
}

#ifdef getenv
#undef getenv
#endif

char *csoundGetEnv(const char *name)
{
    int   i;
    char  *s;

    /* first, try the local database */
    if (csoundEnv_) {
      for (i = 0; i < csoundNumEnvs_; i++) {
        if (strcmp(csoundEnv_[i].name, name) == 0) {
          return (csoundEnv_[i].value);
        }
      }
    }
    /* not found, continue search with system environment variables */
    s = getenv(name);
    /* and return whatever was (or was not) found */
    return s;
}

void csoundReset(void *csound)
{
    IGN(csound);
    SfReset();
    mainRESET();
    csoundIsScorePending_ = 1;
    csoundScoreOffsetSeconds_ = (MYFLT) 0.0;
}

#ifdef INGALLS
  /*
  * Returns a non-zero to the host,
  * which may call csoundCleanup().
  */
void exit(int status)
{
    longjmp(csoundJump_, status +1);
}

int atexit(void (*func)(void))
{
    if (++csoundNumExits_ < csoundMaxExits)
    {
      csoundExitFuncs_[csoundNumExits_] = func;
      return 0;
    }
    else
    {
      return -1;
    }
}
#endif

/* stubs */

void *csoundOpenLibrary(const char *libraryPath)
{
    IGN(libraryPath);
    printf("DO NOT USE THIS INTERFACE\n");
    return NULL;
}

void *csoundCloseLibrary(void *library)
{
    IGN(library);
    printf("DO NOT USE THIS INTERFACE\n");
    return NULL;
}

void *csoundGetLibrarySymbol(void *library, const char *procedureName)
{
    IGN(library);
    IGN(procedureName);
    printf("DO NOT USE THIS INTERFACE\n");
    return NULL;
}

#ifdef __cplusplus
};
#endif

