#ifdef LINUX
#include <dlfcn.h>
#elif defined(WIN32)
#define PUBLIC __declspec(dllexport)
#define LIBRARY_CALL WINAPI
#include <windows.h>
#endif
#include "cs.h"

int csoundLoadExternal(void *csound, const char *libraryPath)
{
#ifdef LINUX
    void    *handle;
    char    *error;
#elif WIN32
    long    handle;
#else
    void    *handle = NULL;
#endif
    void    *(*init)(GLOBALS*);
    long    (*size)(void);
    int     nr_opcodes = 0, length, olength;
    void    *opcodlst_n;

    IGN(csound);
#ifdef LINUX
    handle = dlopen (libraryPath, RTLD_NOW | RTLD_GLOBAL);
#elif WIN32
    handle = (long) LoadLibrary(libraryPath);
#endif
    if (!handle) {
      printf("Failed to load %s\n", libraryPath);
      return (int) CSOUND_ERROR;
    }
#ifdef LINUX
    /* check the number of opcodes, */
    size = (long(*)(void)) dlsym(handle, "opcode_count");
    if ((error = dlerror()) != NULL) {
      /* or the total size of OENTRY structures (deprecated) */
      size = (long(*)(void)) dlsym(handle, "opcode_size");
    }
    else
      nr_opcodes = 1;         /* using newer interface */
    if ((error = dlerror()) != NULL)  {
      printf("Failed to initialise %s: %s", libraryPath, error);
      return (int) CSOUND_ERROR;
    }
    length = (int) (*size)();
    if (!nr_opcodes)
      if ((int) ((length / sizeof(OENTRY)) * sizeof(OENTRY)) != length) {
        printf("%s is not compatible with this version of Csound\n",
               libraryPath);
        return (int) CSOUND_ERROR;
      }
    if (length < 1) {
      printf("%s: invalid opcode count\n", libraryPath);
    }
    /* first check for an init function with the appropriate float type */
#ifdef MYFLT_IS_FLOAT
    init = (void *(*)(GLOBALS*)) dlsym(handle, "opcode_init_32");
#else
    init = (void *(*)(GLOBALS*)) dlsym(handle, "opcode_init_64");
#endif
    /* if that failed, fall back to the generic initialisation function */
    if ((error = dlerror()) != NULL) {
      init = (void *(*)(GLOBALS*)) dlsym(handle, "opcode_init");
    }
    /* still not found */
    if ((error = dlerror()) != NULL)  {
      printf("Failed to initialise %s: %s", libraryPath, error);
      return (int) CSOUND_ERROR;
    }
    opcodlst_n = (*init)(&cglob);
    olength = oplstend - opcodlst;
#elif WIN32
    /* check the number of opcodes, */
    size = (long(*)(void))GetProcAddress((HINSTANCE)handle, "opcode_count");
    if (size == NULL) {
      /* or the total size of OENTRY structures (deprecated) */
      size = (long(*)(void))GetProcAddress((HINSTANCE)handle, "opcode_size");
    }
    else
      nr_opcodes = 1;         /* using newer interface */
    if (size == NULL)  {
      printf("Failed to initialise %s", libraryPath);
      return (int) CSOUND_ERROR;
    }
    length = (int) (*size)();
    if (!nr_opcodes)
      if (((length / sizeof(OENTRY)) * sizeof(OENTRY)) != length) {
        printf("%s is not compatible with this version of Csound\n",
               libraryPath);
        return (int) CSOUND_ERROR;
      }
    if (length < 1) {
      printf("%s: invalid opcode count\n", libraryPath);
    }
    /* first check for an init function with the appropriate float type */
#ifdef MYFLT_IS_FLOAT
    init = (void *(*)(GLOBALS*))
      GetProcAddress((HINSTANCE)handle, "opcode_init_32");
#else
    init = (void *(*)(GLOBALS*))
      GetProcAddress((HINSTANCE)handle, "opcode_init_64");
#endif
    /* if that failed, fall back to the generic initialisation function */
    if (init == NULL) {
      init = (void *(*)(GLOBALS*))
        GetProcAddress((HINSTANCE)handle, "opcode_init");
    }
    /* still not found */
    if (init == NULL)  {
      printf("Failed to initialise %s", libraryPath);
      return (int) CSOUND_ERROR;
    }
#endif
    opcodlst_n = (*init)(&cglob);
    olength = oplstend - opcodlst;
    /* newer interface returns OENTRY** */
    if (nr_opcodes) {
      int i = 0;
      nr_opcodes = length;
      opcodlst = (OENTRY*) mrealloc(opcodlst,
                                    (olength + nr_opcodes) * sizeof(OENTRY));
      oplstend = opcodlst + olength;
      do {
        if (((OENTRY**) opcodlst_n)[i] != NULL) {
          /* clear entry */
          memset(oplstend, 0, sizeof(OENTRY));
          /* copy all the standard fields */
          oplstend->opname    = ((OENTRY**) opcodlst_n)[i]->opname;
          oplstend->dsblksiz  = ((OENTRY**) opcodlst_n)[i]->dsblksiz;
          oplstend->thread    = ((OENTRY**) opcodlst_n)[i]->thread;
          oplstend->outypes   = ((OENTRY**) opcodlst_n)[i]->outypes;
          oplstend->intypes   = ((OENTRY**) opcodlst_n)[i]->intypes;
          oplstend->iopadr    = ((OENTRY**) opcodlst_n)[i]->iopadr;
          oplstend->kopadr    = ((OENTRY**) opcodlst_n)[i]->kopadr;
          oplstend->aopadr    = ((OENTRY**) opcodlst_n)[i]->aopadr;
          oplstend->dopadr    = ((OENTRY**) opcodlst_n)[i]->dopadr;
          oplstend++;
        }
      } while (++i < nr_opcodes);
    }
    else {
      /* old interface: returns OENTRY* */
      opcodlst = (OENTRY*) mrealloc(opcodlst,
                                    olength * sizeof(OENTRY) + length);
      oplstend = opcodlst + olength;
      memcpy(oplstend, (OENTRY*) opcodlst_n, length);
      oplstend += (int) (length / sizeof(OENTRY));
    }
    return (int) CSOUND_SUCCESS;
}

int csoundLoadExternals(void *csound)
{
    char *libname;
    char buffer[256];

    IGN(csound);

    if (CSGLOBAL(oplibs) == NULL) return CSOUND_SUCCESS;
    printf("Loading libraries %s\n", CSGLOBAL(oplibs));
    strcpy(buffer, CSGLOBAL(oplibs));
    libname = strtok(buffer, ",");

    while (libname != NULL) {
      csoundLoadExternal(NULL, libname);
      libname = strtok(NULL, ",");
    }

    return CSOUND_SUCCESS;
}

