/*                                               RTAUDIO.C for WIN32   */

/*  This module is included when RTAUDIO is defined at compile time.
    It provides an interface between Csound realtime record/play calls
    and the device-driver code that controls the actual hardware.
 */
/* RWD modified 15.7.99 - NT-friendly, removed format switch blocks,
   enabling m/c performance eliminated timer from playback code
 */
/* IV - Jan 11 2003: extensively revised code (many parts rewritten), and
   added these features:
     priority control (normal, high, critical)
     variable number of buffers (RWD version was hard-coded to 4)
     sleep time can be controlled for optimal performance
     method for preparing headers can be selected (also for performance)
     code cleaned up
*/
#include "cs.h"
#include "soundio.h"
#undef MYFLT

#include <time.h>
#undef VERSION
#define _WINSOCKAPI_
#ifndef __CYGWIN__
#include <wtypes.h>
#else
#include <windows.h>            /* IV - Oct 26 2002 */
#endif
#include <mmsystem.h>
/* IV - Jan 30 2003 */
#define MAXNUMBUF  (64)

static  HWAVEOUT    outdev;                 /* Device handle */
static  int         numbuf_out;             /* number of buffers */
static  LPSTR       wavbuff[MAXNUMBUF];     /* N buffers */
static  WAVEHDR     wavhdr[MAXNUMBUF];      /* and associated headers */

static  HWAVEIN     indev;                  /* Device handle */
static  int         numbuf_in;              /* number of buffers */
static  LPSTR       wavInbuff[MAXNUMBUF];   /* N buffers */
static  WAVEHDR     wavInhdr[MAXNUMBUF];    /* and associated headers */

static  int    rtout_enabled = 0;   /* RWD 12:2001 use this for playback only */
static  int    rtin_enabled = 0;    /* RWD 12:2001 and this for record only */

static  int    rwd_hdrs = 0;        /* for tuning performance */
static  int    rtwin_sleep = 1;

void RTwavclose(void);

/* RWD.2.98 for GUI Device Selector control */
        int     getWaveOutDevices(void);
        void    getWaveOutName(int, char *);
extern  int     getWantedDevice(void);       /* cwin.cpp */

static  long    inrecs;
static  int     sched_priority = 2;     /* 0: critical, 1: high, 2: normal */

void set_win32_priority(int n)          /* called from argdecode to set */
{                                       /* --sched level */
    if (n < 0) n = 0;
    if (n > 2) n = 2;
    sched_priority = n;
}

void set_win32_sleep_time(int n)        /* called from argdecode to set */
{                                       /* --rt-sleep parameter */
    if (n < 0) n = 0;
    if (n > 10) n = 10;
    rtwin_sleep = n;
}

void set_win32_hdr_mode(int n)          /* called from argdecode to */
{                                       /* enable RWD header mode */
    rwd_hdrs = (n ? 1 : 0);             /* (--rt-hdr) */
}

/* set scheduling priority */

static void set_priority_level(void)
{
    if (sched_priority < 0) return;     /* already set */
    switch (sched_priority) {
    case 0:                     /* critical */
      SetPriorityClass(GetCurrentProcess(), REALTIME_PRIORITY_CLASS);
      SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_HIGHEST);
      break;
    case 1:                     /* high */
      SetPriorityClass(GetCurrentProcess(), HIGH_PRIORITY_CLASS);
      SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_ABOVE_NORMAL);
      break;
    default:                    /* normal */
      SetPriorityClass(GetCurrentProcess(), NORMAL_PRIORITY_CLASS);
      SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_NORMAL);
      break;
    }
    sched_priority = -1;        /* mark as already set */
}

/* print error message for waveInOpen or waveOutOpen and exit */

static void rtopen_error(int n)
{
    char *s = NULL;
    switch (n) {
    case MMSYSERR_BADDEVICEID:
      s = Str(X_295,"ID is out of range");
      break;
    case MMSYSERR_ALLOCATED:
      s = Str(X_190,"already allocated");
      break;
    case MMSYSERR_NOMEM:
      s = Str(X_501,"unable to allocate or lock memory");
      break;
    case WAVERR_BADFORMAT:
      s = Str(X_1360,"unsupported wave format");
      break;
    default:
      s = Str(X_184,"???");
      break;
    }
    printf(Str(X_253,"Error code: %s\n"), s);
    die(Str(X_272,"Failed to open dac"));
}

static void rtopen_select_sample_format(int fmt, float sr, int nchns,
                                        WAVEFORMATEX *wfx)
{
    switch (fmt) {
    case AE_UNCH:
      wfx->wBitsPerSample = 8;
      break;
    case AE_SHORT:
      wfx->wBitsPerSample = 16;
      break;
    case AE_LONG:
      wfx->wBitsPerSample = 32;
      break;
    default:
      err_printf(Str(X_530,"WAVE OUT unknown wave format\n"));
      longjmp(exitjmp, 1);
      break;
    }
    wfx->wFormatTag = WAVE_FORMAT_PCM;
    wfx->nChannels = nchns;
    wfx->nSamplesPerSec = (int) (sr + 0.5f);
    wfx->nAvgBytesPerSec =
      (int) (sr + 0.5f) * nchns * (wfx->wBitsPerSample >> 3);
    wfx->nBlockAlign = nchns * (wfx->wBitsPerSample >> 3);
}

void recopen_(int nchanls, int dsize, float sr, int scale)
                                /* open for audio input */
{
    int           ndev = (int) waveInGetNumDevs(), j, b, bufbytes;
    WAVEINCAPS    caps;
    WAVEFORMATEX  wavform;      /* RWD: I assume Win32s knows about this... */

    IGN(dsize);
    IGN(scale);
    /* device not closed (only possible in reentrant operation) ? */
    if (rtin_enabled) RTwavclose();     /* now it is */
    /* select device */
    if (ndev < 1) die(Str(X_371,"No sound input capabilities"));
    if (ndev > 1) {             /* We ought to allow used to choise but how? */
      printf(Str(X_29,"%d WAVE IN devices found\n"), ndev);
      for (j = 0; j < (int) ndev; j++) {
        waveInGetDevCaps(j, &caps, sizeof(caps));
        printf(Str(X_528,"WAVE IN device %d: %s\n"), j, caps.szPname);
      }
      j = ((int) rtin_dev < ndev ? (int) rtin_dev : 0);
      printf(Str(X_460,"Selecting device %d\n"), j);
      ndev = j;
    }
    else ndev = 0;
    if (waveInGetDevCaps(ndev, &caps, sizeof(WAVEINCAPS)) != MMSYSERR_NOERROR)
      die(Str(X_207,"Cannot get capabilities"));

    /* set sample format */
    rtopen_select_sample_format(CSGLOBAL(O).informat, sr, nchanls, &wavform);

    /* open device */
    if ((b = (int) waveInOpen((LPHWAVEIN) &indev, ndev,
                               &wavform,
                               (DWORD) NULL,
                               (DWORD) NULL, CALLBACK_NULL))
        != MMSYSERR_NOERROR) {
      rtopen_error(b);
      return;
    }
    printf(Str(X_92,"--->WAVE IN DEV.#%d ENABLED  ( %s )\n"),
           ndev, caps.szPname);

    /* set up buffers */
    bufbytes = (CSGLOBAL(O).inbufsamps / nchnls) * (int) wavform.nBlockAlign;
    numbuf_in = CSGLOBAL(O).oMaxLag / (CSGLOBAL(O).inbufsamps / nchnls);
    for (b = 0; b < numbuf_in; b++) {
      WAVEHDR *whdr = &wavInhdr[b];
      wavInbuff[b] = whdr->lpData =
        GlobalLock(GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE, bufbytes));
      memset(wavInbuff[b], 0, bufbytes);
      whdr->dwBufferLength = (DWORD) bufbytes;
      whdr->dwUser = 0;
      whdr->dwFlags = WHDR_DONE;
      whdr->dwLoops = 1;
      waveInPrepareHeader(indev, whdr, sizeof(WAVEHDR));
    }
    printf("WAVE IN: using %d buffers of %d bytes\n", numbuf_in, bufbytes);
    waveInStart(indev);
    rtin_enabled = 1;
    inrecs = 0L;                        /* IV - Jan 15 2003 */
    /* set scheduling priority */
    set_priority_level();
}

void playopen_(int nchanls, int dsize, float sr, int scale)
                                /* open for audio output */
{
    int           ndev = (int) waveOutGetNumDevs(), j, b, bufbytes;
    WAVEOUTCAPS   caps;
    WAVEFORMATEX  wavform;      /* RWD: I assume Win32s knows about this... */

    IGN(dsize);
    IGN(scale);
    /* device not closed (only possible in reentrant operation) ? */
    if (rtout_enabled) RTwavclose();    /* now it is */
    /* select device */
    if (ndev < 1) die(Str(X_370,"No sound capabilities"));
    if (ndev > 1) {             /* We ought to allow used to choise but how? */
      printf(Str(X_30,"%d WAVE OUT devices found\n"), ndev);
      for (j = 0; j < ndev; j++) {
        waveOutGetDevCaps(j, &caps, sizeof(caps));
        printf(Str(X_529,"WAVE OUT device %d: %s\n"), j, caps.szPname);
      }
      j = ((int) rtout_dev < ndev ? (int) rtout_dev : 0);
      printf(Str(X_460,"Selecting device %d\n"), j);
      ndev = j;
    }
    else ndev = 0;
    if (waveOutGetDevCaps(ndev, &caps, sizeof(WAVEOUTCAPS)) != MMSYSERR_NOERROR)
      die(Str(X_207,"Cannot get capabilities"));

    /* set sample format */
    rtopen_select_sample_format(CSGLOBAL(O).outformat, sr, nchanls, &wavform);

    /* open device */
    if ((b = (int) waveOutOpen((LPHWAVEOUT) &outdev, ndev,
                               &wavform,
                               (DWORD) NULL,
                               (DWORD) NULL, CALLBACK_NULL))
        != MMSYSERR_NOERROR) {
      rtopen_error(b);
      return;
    }
    printf(Str(X_93,"--->WAVE OUT DEV.#%d ENABLED  ( %s )\n"),
           ndev, caps.szPname);

    /* set up buffers */
    bufbytes = (CSGLOBAL(O).outbufsamps / nchnls) * (int) wavform.nBlockAlign;
    numbuf_out = CSGLOBAL(O).oMaxLag / (CSGLOBAL(O).outbufsamps / nchnls);
    for (b = 0; b < numbuf_out; b++) {
      WAVEHDR *whdr = &wavhdr[b];
      wavbuff[b] = wavhdr[b].lpData =
        GlobalLock(GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE, bufbytes));
      memset(wavbuff[b], 0, bufbytes);
      whdr->dwBufferLength = bufbytes;
      whdr->dwUser = 0;
      whdr->dwFlags = WHDR_DONE;
      whdr->dwLoops = (rwd_hdrs ? 0 : 1);
    }
    printf("WAVE OUT: using %d buffers of %d bytes\n", numbuf_out, bufbytes);
    rtout_enabled = 1;
    /* set scheduling priority */
    set_priority_level();
    /* prepare headers now if enabled */
    if (!rwd_hdrs)
      for (b = 0; b < numbuf_out; b++)
        waveOutPrepareHeader(outdev, &wavhdr[b], sizeof(WAVEHDR));
}

int rtrecord_(char *inbuf, int nbytes) /* get samples from ADC */
{
    int   ndx = (int) inrecs % numbuf_in;
    int   prvndx = (ndx ? (ndx - 1) : (numbuf_in - 1));
    volatile DWORD* dwflg = &(wavInhdr[ndx].dwFlags);
    if (inrecs >= numbuf_in) {
      if (rtwin_sleep)                          /* if input too fast */
        while ((*dwflg & WHDR_DONE) == 0) Sleep(rtwin_sleep);
      else
        while ((*dwflg & WHDR_DONE) == 0);
    }
    /* copy data */
    memcpy(inbuf, wavInbuff[ndx], nbytes);
    /* prepare new buffer */
    if (rwd_hdrs) {
      waveInPrepareHeader(indev, &wavInhdr[prvndx], sizeof(WAVEHDR));
      wavInhdr[ndx].dwBufferLength = nbytes;    /* ??? */
    }
    waveInAddBuffer(indev, &wavInhdr[prvndx], sizeof(WAVEHDR));
    /* next buffer */
    inrecs++;
    return(nbytes);
}

void rtplay_(char *outbuf, int nbytes)          /* put samples to DAC  */
  /* N.B. This routine serves as a THROTTLE in Csound Realtime Performance, */
  /* delaying the actual writes and return until the hardware output buffer */
  /* passes a sample-specific THRESHOLD.  If the I/O BLOCKING functionality */
  /* is implemented ACCURATELY by the vendor-supplied audio-library write,  */
  /* that is sufficient.  Otherwise, requires some kind of IOCTL from here. */
  /* This functionality is IMPORTANT when other realtime I/O is occurring,  */
  /* such as when external MIDI data is being collected from a serial port. */
  /* Since Csound polls for MIDI input at the software synthesis K-rate     */
  /* (the resolution of all software-synthesized events), the user can      */
  /* eliminate MIDI jitter by requesting that both be made synchronous with */
  /* the above audio I/O blocks, i.e. by setting -b to some 1 or 2 K-prds.  */
{
    int   ndx = (int) nrecs % numbuf_out;
    volatile DWORD* dwflg = &(wavhdr[ndx].dwFlags);
    /*RWD.7.99: only test WHDR_DONE when all blocks have */
    /* been used for the first time! */
    if (nrecs >= numbuf_out) {
      if (rtwin_sleep)                          /* if input too fast */
        while ((*dwflg & WHDR_DONE) == 0) Sleep(rtwin_sleep);
      else
        while ((*dwflg & WHDR_DONE) == 0);
      if (rwd_hdrs)
        waveOutUnprepareHeader(outdev, &wavhdr[ndx], sizeof(WAVEHDR));
    }
    /* copy data */
    memcpy(wavbuff[ndx], outbuf, nbytes);
    if (rwd_hdrs) {
      wavhdr[ndx].dwBufferLength = nbytes;
      /* RWD.7.99 ...but we do want to do it here */
      waveOutPrepareHeader(outdev, &wavhdr[ndx], sizeof(WAVEHDR));
    }
    waveOutWrite(outdev, &wavhdr[ndx], sizeof(WAVEHDR));
    /* next buffer */
    nrecs++;
}

/* RWD: extra func for use in cwin_exit(): to close device on premature exit, */
/* and make all this reentry-friendly*/

void RTwavclose(void)
{
    /* close output device */
    if (rtout_enabled) {
      waveOutReset(outdev);
      waveOutClose(outdev);
      rtout_enabled = 0;        /* no longer active */
    }
    /* same for input */
    if (rtin_enabled) {
      waveInStop(indev);
      waveInReset(indev);
      waveInClose(indev);
      rtin_enabled = 0;
    }
    /* reset scheduling priority to normal */
    set_win32_priority(2);
    set_priority_level();
    set_win32_priority(2);
    /* reset performance tuning parameters */
    rwd_hdrs = 0;
    rtwin_sleep = 1;
}

void rtclose_(void)              /* close the I/O device entirely  */
{                               /* called only when both complete */
    /* RWD.7.99: just wait for the last block to be DONE, and we're done! */
    /* RWD 12:2001 test we are doing this! */
    if (rtout_enabled) {
      int lastblock = (nrecs + (numbuf_out - 1)) % numbuf_out;
      while ((wavhdr[lastblock].dwFlags & WHDR_DONE) == 0)
        Sleep(1);
    }
    /* RWD 12:2001 and test here... */
    RTwavclose();
    if (CSGLOBAL(O).Linein) {
#ifdef PIPES
#ifdef __CYGWIN__
      extern int _pclose(FILE*);
#endif
      if (CSGLOBAL(O).Linename[0] == '|' && Linepipe) {
        _pclose(Linepipe);
        Linepipe = NULL;
      }
      else
#endif
      if (strcmp(CSGLOBAL(O).Linename, "stdin") && Linefd >= 0) {
        close(Linefd);
        Linefd = -1;
      }
    }
}

/* RWD.2.98 new wrapper funcs for GUI selection of WaveOut Device */
int getWaveOutDevices(void)
{
    return waveOutGetNumDevs();
}

void getWaveOutName(int dev, char *name)
{
    WAVEOUTCAPS caps;
    if (name == NULL) return;
    waveOutGetDevCaps(dev, &caps, sizeof(caps));
    strcpy(name, caps.szPname);
}

